/*
popclk.c - Population Clock
Version 1.0.0
Copyright (C) 2001-2004  dondalah@ripco.com (Dondalah)

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to:

    Free Software Foundation, Inc.
    59 Temple Place - Suite 330
    Boston, MA  02111-1307, USA.
*/
#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include "popclk.h"

#define ID_TIMER 1
#define RELTIME 0
#define ABSTIME 1
#define ANNRATE 72898133.0

int iErrSw,iTimeSw,iWait,iExpose;
time_t iOrigTime,iPrevTime;
int iInitTicks;
int cxClient,cyClient;
int cxMiddle,cyMiddle;
int cxChar,cyChar,cxCaps;
double dAnnRate,dTicRate;
double dTotal,dPrevTotal,dAccumTotal;
char szTitle[128];
char szPrevTitle[128];

/*******************************************************************
char szDbg1[80];
char szDbg2[80];
char szDbg3[80];
*******************************************************************/

LRESULT CALLBACK WndProc      (HWND, UINT, WPARAM, LPARAM);
VOID    CALLBACK TimerProc    (HWND, UINT, UINT,   DWORD);
BOOL    CALLBACK ChgDlgProc   (HWND, UINT, WPARAM, LPARAM);
BOOL    CALLBACK TitleDlgProc (HWND, UINT, WPARAM, LPARAM);

static char szAppName[] = "PopClk";

int WINAPI WinMain (HINSTANCE hInstance,
                    HINSTANCE hPrevInstance,
                    PSTR szCmdLine,
                    int iCmdShow)
   {
   HWND hwnd;
   MSG msg;
   WNDCLASSEX wndclass;

   wndclass.cbSize         = sizeof(wndclass);
   wndclass.style          = CS_HREDRAW | CS_VREDRAW;
   wndclass.lpfnWndProc    = WndProc;
   wndclass.cbClsExtra     = 0;
   wndclass.cbWndExtra     = 0;
   wndclass.hInstance      = hInstance;
   wndclass.hIcon          = LoadIcon (hInstance, szAppName);
   wndclass.hCursor        = LoadCursor (NULL, IDC_ARROW);
   wndclass.hbrBackground  = (HBRUSH)  GetStockObject (WHITE_BRUSH);
   wndclass.lpszMenuName   = szAppName;
   wndclass.lpszClassName  = szAppName;
   wndclass.hIconSm        = LoadIcon (hInstance, szAppName);

   RegisterClassEx (&wndclass);

   hwnd = CreateWindow (szAppName,  // Window Class Name
      "Population Clock",           // Window Caption
      WS_OVERLAPPEDWINDOW,          // Window Style
      60,                           // initial x position
      60,                           // initial y position
      323,                          // initial x size
      200,                          // initial y size
      NULL,                         // parent window handle
      NULL,                         // window menu handle
      hInstance,                    // program instance handle
      NULL);                        // creation parameters

   { /* set timer */
   int iResult;
   iResult = SetTimer(hwnd, ID_TIMER, 10,
                     (TIMERPROC) TimerProc);
   if (!iResult)
      {
      MessageBox(hwnd,
         "Too many clocks or timers!",
         "Population Clock",
         MB_ICONEXCLAMATION | MB_OK);
      return FALSE;
      } /* too many timers */
   } /* set timer */

   ShowWindow (hwnd, iCmdShow);
   UpdateWindow (hwnd);

   while (GetMessage (&msg, NULL, 0, 0))
      {
      TranslateMessage (&msg);
      DispatchMessage (&msg);
      } // while (GetMessage (&msg, NULL, 0, 0))

   return msg.wParam;
   } /* WinMain */

LRESULT CALLBACK WndProc (HWND hwnd,
                          UINT iMsg,
                          WPARAM wParam,
                          LPARAM lParam)
   {
   static HINSTANCE hInstance;
   HMENU hMenu;
   TEXTMETRIC tm;
   HDC hdc;

   switch (iMsg)
      {
      case WM_CREATE:

         {
         int iLength;
         double dSecPerYear;
         double dTicPerYear;
         dAnnRate    = ANNRATE;
         dSecPerYear = 3600.0 * 24.0 * 365.25;
         dTicPerYear = dSecPerYear * 1000;
         dTicRate    = dAnnRate / dTicPerYear;
         iTimeSw     = RELTIME;
         time(&iOrigTime);
         iPrevTime   = iOrigTime;
         iInitTicks  = clock();
         dAccumTotal = dTotal = 0.0;
         dPrevTotal = -999999999.0;
         iLength = GetWindowText(hwnd,szTitle,120);
         strcpy(szPrevTitle,szTitle);
         hMenu = GetMenu(hwnd);
         CheckMenuItem(hMenu,
            IDM_RELATIVE,MF_CHECKED);
         hdc = GetDC(hwnd);
         GetTextMetrics(hdc,&tm);
         cxChar = tm.tmAveCharWidth;
         cxCaps = (tm.tmPitchAndFamily & 1 ? 3 : 2)
            * (cxChar >> 1);
         cyChar = tm.tmHeight + tm.tmExternalLeading;
         ReleaseDC(hwnd,hdc);
         iWait = iExpose = 0;

         return 0;
         } /* create */

      case WM_SIZE:
         cxClient = LOWORD (lParam);
         cyClient = HIWORD (lParam);
         cxMiddle = cxClient >> 1;
         cyMiddle = cyClient >> 1;
         if (iWait)
            {
            iWait = 0;
            InvalidateRect(hwnd,NULL,TRUE);
            } /* if waiting for copyright screen */
         return 0;

      case WM_COMMAND:

         hMenu = GetMenu(hwnd);
         switch (LOWORD (wParam))
            {
            case IDM_EXIT:
               PostMessage(hwnd,WM_CLOSE,0,0L);
               return 0;

            case IDM_RESET:
               time(&iOrigTime);
               iPrevTime   = iOrigTime;
               iInitTicks  = clock();
               dAccumTotal = dTotal = 0.0;
               dPrevTotal = -999999999.0;
               return 0;

            case IDM_COPYING:
               { /* copying popclk */
               int cxOfst,cyOfst;
               char szTemp[128];
               hdc = GetDC(hwnd);

               cxOfst = cxMiddle - (cxChar << 3);
               cyOfst = cyMiddle - cyChar - cyChar;
               sprintf(szTemp, "                               ");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxMiddle - (cxChar << 2);
               cyOfst = cyMiddle;
               sprintf(szTemp, "                               ");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));

               cxOfst = cxChar + 4;
               cyOfst = cyMiddle - (cyChar * 4);
               sprintf(szTemp, "Popclk.c Version 1.0.0 "
                  "- Population Clock");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxChar + 4;
               cyOfst = cyMiddle - (cyChar * 3);
               sprintf(szTemp, "Copyright (C) 2001-2004");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxChar + 4;
               cyOfst = cyMiddle - (cyChar * 2);
               sprintf(szTemp, "by dondalah@ripco.com (Dondalah)");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxChar + 4;
               cyOfst = cyMiddle - (cyChar * 1);
               sprintf(szTemp, "Popclk comes with");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxChar + 4;
               cyOfst = cyMiddle - (cyChar * 0);
               sprintf(szTemp, "ABSOLUTELY NO WARRANTY");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxChar + 4;
               cyOfst = cyMiddle + (cyChar * 1);
               sprintf(szTemp, "For details, read the "
                  "GNU Public License at");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxChar + 4;
               cyOfst = cyMiddle + (cyChar * 2);  
               sprintf(szTemp, "http://www.gnu.org/"
                  "copyleft/gpl.html");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               cxOfst = cxChar + 4;
               cyOfst = cyMiddle + (cyChar * 3);
               sprintf(szTemp, "Press any key to "
                  "continue.");
               TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
               ReleaseDC(hwnd,hdc);
               iWait = 1;
               } /* copying popclk */
               return 0;

            case IDM_ABOUT:
               MessageBox(hwnd,
                  "Population Clock 1.0.0",
                  szAppName,
                  MB_ICONINFORMATION | MB_OK);
               iExpose = 1;
               return 0;

            case IDM_SEC_CHG:
               {
               double dSecPerChg;
               char svMsg[256];
               /* 1000 ticks per second */
               dSecPerChg = 1.0 / fabs(dTicRate) / 1000.0;
               sprintf(svMsg,
                  "The total changes every %30.15f seconds",
                  dSecPerChg);
               MessageBox(hwnd,
                  svMsg,
                  szAppName,
                  MB_ICONINFORMATION | MB_OK);
               iExpose = 1;
               return 0;
               } /* case IDM_SEC_CHG */

            case IDM_CHANGE:
               {
               int iResult;
               iResult = DialogBox(hInstance,"ChangeBox",
                         hwnd,ChgDlgProc);
               if (iResult == -1)
                  {
                  MessageBox(hwnd,"Error","Change Rate",
                     MB_ICONERROR|MB_OK);
                  return TRUE;
                  } /* if err */
               return TRUE;
               } /* IDM_CHANGE */

            case IDM_TITLE:
               {
               int iResult;
               iResult = DialogBox(hInstance,"TitleBox",
                         hwnd,TitleDlgProc);
               if (iResult == -1)
                  {
                  MessageBox(hwnd,"Error",
                     "Change Caption",
                     MB_ICONERROR|MB_OK);
                  return TRUE;
                  } /* if err */
               return TRUE;
               } /* IDM_TITLE */

            case IDM_RELATIVE:
               {
               iTimeSw = RELTIME;
               CheckMenuItem(hMenu,
                  IDM_ABSOLUTE,MF_UNCHECKED);
               CheckMenuItem(hMenu,
                  IDM_RELATIVE,MF_CHECKED);
               return TRUE;
               } /* IDM_RELATIVE */

            case IDM_ABSOLUTE:
               {
               iTimeSw = ABSTIME;
               CheckMenuItem(hMenu,
                  IDM_RELATIVE,MF_UNCHECKED);
               CheckMenuItem(hMenu,
                  IDM_ABSOLUTE,MF_CHECKED);
               return TRUE;
               } /* IDM_ABSOLUTE */

            } /* switch (LOWORD (wParam)) */
         break;

      case WM_LBUTTONDOWN:

         if (iWait)
            {
            iWait = 0;
            InvalidateRect(hwnd,NULL,TRUE);
            iExpose = 1;
            } /* if waiting for copy screen */
         else
            {
            PostMessage(hwnd,WM_CLOSE,0,0L);
            } /* not waiting for copy screen */

         return 0;

      case WM_KEYDOWN:

         if (iWait)
            {
            iWait = 0;
            InvalidateRect(hwnd,NULL,TRUE);
            iExpose = 1;
            }

         if (wParam == 'Q')
            {
            PostMessage(hwnd,WM_CLOSE,0,0L);
            } /* if quit */

         else if (wParam == 'R')
            {
            time(&iOrigTime);
            iPrevTime = iOrigTime;
            iInitTicks = clock();
            dTotal = 0.0;
            dPrevTotal = -999999999.0;
            } /* if reset */

         else if (wParam == 'T')
            {
            if (iTimeSw == ABSTIME)
               {
               iTimeSw = RELTIME;
               } /* clock time to relative time */
            else
               {
               iTimeSw = ABSTIME;
               } /* relative time to clock time */
            } /* if reset */

         return 0;

      case WM_DESTROY:
         PostQuitMessage (0);
         return 0;

      } /* switch (iMsg) */

   return DefWindowProc (hwnd, iMsg, wParam, lParam);
   } /* WndProc */

void CALLBACK TimerProc (HWND hwnd,
                         UINT iMsg,
                         UINT iTimerID,
                         DWORD dwTime)
   {
   HDC hdc;
   RECT rect;

   GetClientRect (hwnd, &rect);
   hdc = GetDC(hwnd);

   if (iWait)
      {
      /* waiting for copyright screen */
      ReleaseDC(hwnd,hdc);
      return;
      } /* if waiting for keypress */

   { /* calculate new total at each tick */
   int iCurrTicks,iTotTicks;
   iCurrTicks = clock();
   if (iCurrTicks < iInitTicks)
      {
      iInitTicks = iCurrTicks;
      dAccumTotal += dTotal;
      dTotal = 0.0;
      dPrevTotal = -999999999.0;
      } /* if wrap around */
   iTotTicks = iCurrTicks - iInitTicks;
   dTotal = dAccumTotal
      + (dTicRate * (double) iTotTicks);
   } /* calculate new total at each tick */

   { /* paint output */
   int iResult;
   char szTemp[128];
   struct tm tstr;
   struct tm *ts;
   time_t iCurrTime;
   int iElap,iTimeDiff;
   int cxOfst,cyOfst;
   if (fabs(dTotal - dPrevTotal) > 1.0
      || iExpose)
      {
      cxOfst = cxMiddle - (cxChar << 3);
      cyOfst = cyMiddle - cyChar - cyChar;
      sprintf(szTemp, "                               ");
      TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
      SetTextAlign(hdc,TA_RIGHT | TA_TOP);
      cxOfst = cxMiddle - (cxChar << 3) + (cxChar * 12);
      sprintf(szTemp, "%15.0f", dTotal);
      TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
      SetTextAlign(hdc,TA_LEFT | TA_TOP);
      dPrevTotal = dTotal;
      } /* if print new total */
   time(&iCurrTime);
   iTimeDiff = iCurrTime - iPrevTime;
   iElap = iCurrTime - iOrigTime;
   if (iTimeDiff || iExpose)
      {
      if (iTimeSw == RELTIME)
         {
         ts = &tstr;
         ts->tm_hour = iElap / 3600;
         ts->tm_min  = (iElap % 3600) / 60;
         ts->tm_sec  = iElap % 60;
         } /* relative time */
      else
         {
         ts = localtime(&iCurrTime);
         } /* absolute time */
      cxOfst = cxMiddle - (cxChar << 2);
      cyOfst = cyMiddle;
      sprintf(szTemp, "                               ");
      TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
      sprintf(szTemp, "%02d:%02d:%02d",
         ts->tm_hour, ts->tm_min, ts->tm_sec);
      TextOut(hdc,cxOfst,cyOfst,szTemp,strlen(szTemp));
      iPrevTime = iCurrTime;
      iResult = strcmp(szTitle,szPrevTitle);
      if (iResult)
         {
         SetWindowText(hwnd,szTitle);
         strcpy(szPrevTitle,szTitle);
         } /* if title change */
      } /* if at least 1 sec elapsed */
   } /* paint output */

   iExpose = 0;
   ReleaseDC(hwnd, hdc);
   } /* TimerProc */

BOOL CALLBACK ChgDlgProc (HWND hdlg, UINT iMsg,
                           WPARAM wParam, LPARAM lParam)
   {
   static HWND hCtrlBlock;
   static HWND hwndEdit;
   char szAnnRate[32];
   switch (iMsg)
      {
      case WM_INITDIALOG:
         return TRUE;

      case WM_SETFOCUS:
         /* can't put messagebox here */
         /* will cause endless wm_setfocus calls */
         SetFocus (hwndEdit);
	 sprintf(szAnnRate, "%.0f", ANNRATE);
         SetWindowText(hdlg,szAnnRate);
         return 0;

      case WM_COMMAND:
         switch (LOWORD (wParam))
            {
            case IDOK:
               {
               int iResult;
               double dPrevAnnRate;
               double dSecPerYear;
               double dTicPerYear,dSgn;
               char szStr[128];
               char szDbg[128];
               dPrevAnnRate = dAnnRate;
               /* iResult is length of szStr or -1 */
               iResult = GetDlgItemText(hdlg, IDC_EDITBOX,
                         (LPTSTR) szStr, 100);
               if (iResult == -1)
                  {
                  MessageBox(hdlg,"Dialog Error",
                     "New Annual Rate",
                     MB_ICONERROR|MB_OK);
                  } /* error */
               else if (!iResult)
                  {
                  } /* no change */
               else
                  {
                  char *p,*q;
                  p = szStr;
                  q = p + strlen(szStr);
                  dAnnRate = 0.0;
                  if (*p == '-')
                     {
                     dSgn = -1.0;
                     p++;
                     } /* if prefix sign */
                  else dSgn = 1.0;
                  while (p < q)
                     {
                     if (isdigit(*p))
                        {
                        dAnnRate = (dAnnRate * 10.0)
                           + (double) (*p++ - '0');
                        } /* add digit */
                     else
                        {
                        sprintf(szDbg,
                           "Not an integer: %s",
                           szStr);
                        MessageBox(hdlg,
                           szDbg,
                           "Annual Rate Error",
                           MB_ICONERROR|MB_OK);
                        dAnnRate = 0.0;
                        break;
                        } /* add digit */
                     } /* for each digit */
                  if (dAnnRate == 0.0)
                     {
                     dAnnRate = dPrevAnnRate;
                     EndDialog(hdlg,TRUE);
                     return TRUE;
                     } /* if new rate is zero */
                       /* treat as cancel */
                  if (dAnnRate > 999999999999999.0)
                     {
                     MessageBox(hdlg,
                        "Rate is -999999999999999 to 999999999999999",
                        "Annual Rate Error",
                        MB_ICONERROR|MB_OK);
                     dAnnRate = dPrevAnnRate;
                     } /* rate too large */
                  dSecPerYear = 3600.0 * 24.0 * 365.25;
                  dTicPerYear = dSecPerYear * 1000;
                  dTicRate    = dAnnRate / dTicPerYear;
                  if (dAnnRate >= 1.0
                     && dAnnRate < 10000000.0)
                     {
                     char svSlowMsg[256];
                     sprintf(svSlowMsg,
                        "The total will change every "
                        "%9.1f seconds",
                        1.0 / dTicRate / 1000.0);
                        /* inverse of rate change per second */
                     MessageBox(hdlg,
                        svSlowMsg,
                        "Annual Rate Change",
                        MB_ICONWARNING|MB_OK);
                     } /* slow rate */
                  dTicRate   *= dSgn;
                  dTotal = 0.0;
                  dPrevTotal = -999999999.0;
                  iInitTicks = clock();
                  time(&iOrigTime);
                  iPrevTime = iOrigTime;
                  } /* if got text */
               EndDialog(hdlg,TRUE);
               return TRUE;
               }
            case IDCANCEL:
               EndDialog(hdlg,FALSE);
               // SendMessage(hdlg,WM_CLOSE,0,0L);
               // MessageBox(hdlg,"after enddlg",
                  // "chgdlg,COMMAND,IDCANCEL",MB_OK);
               return TRUE;
            } /* switch (LOWORD (wParam)) */
         return FALSE;

      case WM_CLOSE:
         EndDialog(hdlg,IDOK);
         return TRUE;
      } /* switch (iMsg) */

   return FALSE;
   } /* ChgDlgProc */

BOOL CALLBACK TitleDlgProc (HWND hdlg, UINT iMsg,
                            WPARAM wParam, LPARAM lParam)
   {
   static HWND hCtrlBlock;
   static HWND hwndEdit;
   char szAnnRate[32];
   switch (iMsg)
      {
      case WM_INITDIALOG:
         return TRUE;

      case WM_SETFOCUS:
         /* can't put messagebox here */
         /* will cause endless wm_setfocus calls */
         SetFocus (hwndEdit);
	 sprintf(szAnnRate, "%.0f", ANNRATE);
         SetWindowText(hdlg,szAnnRate);
         return 0;

      case WM_COMMAND:
         switch (LOWORD (wParam))
            {
            case IDOK:
               {
               int iResult;
               char szStr[128];
               char szDbg[128];
               /* iResult is length of szStr or -1 */
               iResult = GetDlgItemText(hdlg, IDC_EDITBOX,
                         (LPTSTR) szStr, 100);
               if (iResult == -1)
                  {
                  MessageBox(hdlg,"Dialog Error",
                     "New Caption",
                     MB_ICONERROR|MB_OK);
                  } /* error */
               else if (!iResult)
                  {
                  } /* no change */
               else if (iResult < 30)
                  {
                  strcpy(szTitle,szStr);
                  } /* if got text */
               else
                  {
                  sprintf(szDbg,
                     "%s",
                     szStr);
                  MessageBox(hdlg,
                     szDbg,
                     "Title too long",
                     MB_ICONERROR|MB_OK);
                  } /* too long */
               EndDialog(hdlg,TRUE);
               return TRUE;
               }

            case IDCANCEL:
               EndDialog(hdlg,FALSE);
               // MessageBox(hdlg,"after enddlg",
                  // "chgdlg,COMMAND,IDCANCEL",MB_OK);
               return TRUE;
            } /* switch (LOWORD (wParam)) */
         return FALSE;

      case WM_CLOSE:
         EndDialog(hdlg,IDOK);
         return TRUE;
      } /* switch (iMsg) */

   return FALSE;
   } /* ChgDlgProc */
